# As usual, the obligatory warning that you will get lost in this code.

import math
import random
import pyglet
import const


class GameWindow(pyglet.window.Window):

    def __init__(self):
        super(GameWindow, self).__init__(const.WIDTH, const.HEIGHT,
                                         const.TITLE)
        self.current_window = Intro(self)
        try:
            player = pyglet.media.Player()
            player.queue(pyglet.resource.media('MTA.mp3', streaming=True))
            player.eos_action = player.EOS_LOOP
            player.play()
        except:
            print 'Unable to load music.'
    
    def start_game(self):
        self.current_window = Game(self)

    def back_to_intro(self):
        self.current_window = Intro(self)


class Intro:

    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        bg_image = (pyglet.image.SolidColorImagePattern((0,0,50,255)).
                    create_image(const.WIDTH, const.HEIGHT))
        self.bg = pyglet.sprite.Sprite(bg_image, 0, 0,
                                       batch=self.batch)
        self.window.push_handlers(self.on_draw, self.on_key_press)
        title_text = 'The Strange Sine Orbs'
        instr_text = 'Z, X and C to toggle the "wobbling" of the top, middle and bottom orbs respectively.'
        instr2_text = '(That is, you can stop the "wobbling" by pressing the button again.)'
        instr3_text = 'Catch the charges and bring them to the right. Do not let enemies reach the left.'
        instr4_text = 'The lines predict your movement.'
        start_text = 'Press Z to start.'
        pyweek_text = 'Made for Pyweek #10 under theme "Wibbly-wobbly".'
        self.title_label = pyglet.text.Label(title_text, batch=self.batch,
            x=const.WIDTH / 2, y=3 * const.HEIGHT / 4, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.instr_label = pyglet.text.Label(instr_text, batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=14,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.instr2_label = pyglet.text.Label(instr2_text, batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2 - 30, font_size=14,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.instr3_label = pyglet.text.Label(instr3_text, batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2 - 60, font_size=14,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.instr4_label = pyglet.text.Label(instr4_text, batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2 - 90, font_size=14,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.start_label = pyglet.text.Label(start_text, batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 4 - 30, font_size=20,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.start_label = pyglet.text.Label(pyweek_text, batch=self.batch,
            x=30, y=30, font_size=10,
            anchor_x='left', anchor_y='center', color=(255,255,255,255))

    def on_draw(self):
        self.window.clear()
        self.batch.draw()

    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.Z:
            self.window.pop_handlers()
            self.window.start_game()


class Game:

    def __init__(self, window):
        self.window = window

        self.paused = False
        self.paused_label = None
        self.end_label = None

        self.batch = pyglet.graphics.Batch()
        self.bg_batch = pyglet.graphics.Batch()

        bg_image = (pyglet.image.SolidColorImagePattern((0,0,50,255)).
                    create_image(const.WIDTH, const.HEIGHT))
        self.bg = pyglet.sprite.Sprite(bg_image, 0, 0,
                                       batch=self.bg_batch)

        self.life_mgr = LifeManager(self.batch)

        self.bullet_gen = BulletGenerator(self.batch, self.life_mgr)
        self.enemy_gen = EnemyGenerator(self.batch, self.life_mgr)
        self.charge_gen = ChargeGenerator(self.batch)

        self.collision_mgr = CollisionManager(self.bullet_gen.bullets,
                                              self.enemy_gen.enemies,
                                              self.charge_gen.charges)

        self.window.push_handlers(self.on_draw, self.on_key_press)
        pyglet.clock.schedule_interval(self.update, const.INV_FPS)

    def finish(self):
        self.window.pop_handlers()
        pyglet.clock.unschedule(self.update)

    def pause(self):
        self.paused = True
        self.paused_label = pyglet.text.Label('Paused', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))

    def unpause(self):
        self.paused = False
        self.paused_label.delete()
        self.paused_label = None

    def on_draw(self):
        self.window.clear()
        self.bg_batch.draw()
        self.bullet_gen.draw()
        self.batch.draw()

    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.P:
            if not self.paused: self.pause()
            else: self.unpause()
        elif symbol == pyglet.window.key.Z:
            self.bullet_gen.toggle_wobbling(2)
        elif symbol == pyglet.window.key.X:
            self.bullet_gen.toggle_wobbling(1)
        elif symbol == pyglet.window.key.C:
            self.bullet_gen.toggle_wobbling(0)
        elif symbol == pyglet.window.key.E:
            self.bullet_gen.toggle_wobbling(2)
        elif symbol == pyglet.window.key.D:
            self.bullet_gen.toggle_wobbling(1)
        elif symbol == pyglet.window.key.R:
            self.finish()
            self.window.back_to_intro()
    
    def update(self, dt):
        if self.paused: return
        dt = const.INV_FPS # this is bad, but I think I need this to predict
        self.bullet_gen.update(dt)
        self.enemy_gen.update(dt)
        self.charge_gen.update(dt)
        self.collision_mgr.update(dt)

        if self.life_mgr.end_code > 0:
            self.win()
        elif self.life_mgr.end_code < 0:
            self.lose()

    def win(self):
        self.end_label = pyglet.text.Label('Awesome! You beat the enemy!', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.paused = True

    def lose(self):
        self.end_label = pyglet.text.Label('The enemy defeats you!', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2 + 40, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        pyglet.text.Label('Play again! Press R to restart.', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2 - 40, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))
        self.paused = True


class Bullet(pyglet.sprite.Sprite):

    def __init__(self, batch, pos):
        self.orig_image = pyglet.resource.image('bullet.png')
        self.orig_image.anchor_x = 6
        self.orig_image.anchor_y = 6

        image_charged = pyglet.resource.image('bullet-anim.png')
        img_seq = pyglet.image.ImageGrid(image_charged, 3, 4)
        self.charge1 = pyglet.image.Animation.from_image_sequence(
            img_seq[8:12], 0.1, True)
        self.charge2 = pyglet.image.Animation.from_image_sequence(
            img_seq[4:8], 0.1, True)
        self.charge3 = pyglet.image.Animation.from_image_sequence(
            img_seq[0:4], 0.1, True)
        for frame in self.charge1.frames:
            frame.image.anchor_x = frame.image.anchor_y = 6
        self.charge1.anchor_x = self.charge1.anchor_y = 6
        for frame in self.charge2.frames:
            frame.image.anchor_x = frame.image.anchor_y = 6
        self.charge2.anchor_x = self.charge2.anchor_y = 6
        for frame in self.charge3.frames:
            frame.image.anchor_x = frame.image.anchor_y = 6
        self.charge3.anchor_x = self.charge3.anchor_y = 6
        super(Bullet, self).__init__(self.orig_image, batch=batch)

        self.x = pos[0]
        self.y = pos[1]

        self.init_vel = 30
        self.vel = [self.init_vel, 0]

        self.wobbling = False
        self.amplitude = 0
        self.amplitude_accel = 20
        self.amplitude_deccel = 25
        self.amp_limit = 40

        self.angle = 0

        self.path_batch = pyglet.graphics.Batch()
        self.other_path_batch = pyglet.graphics.Batch()

        self.current_path = None
        self.path_points = []
        self.path_color = (140, 200, 180)
        self.other_path = None
        self.other_color = (200, 120, 150)
        self.ok_color = (30, 30, 90)

        self.charges = 0
        
        self.done = False

    def reset(self, pos):
        self.x = pos[0]
        self.y = pos[1]

        self.vel = [self.init_vel, 0]

        self.wobbling = False
        self.amplitude = 0

        self.angle = 0

        self.path_batch = pyglet.graphics.Batch()
        self.other_path_batch = pyglet.graphics.Batch()

        self.current_path = None
        self.path_points = []
        self.other_path = None

        self.charges = 0
        self.image = self.orig_image
        
        self.done = False       

    def start_wobbling(self):
        self.wobbling = True
        self.predict(True)

    def stop_wobbling(self):
        self.wobbling = False
        self.amplitude *= 0.5
        self.predict(False)

    def update(self, dt):
        if self.amplitude > const.EPS:
            self.angle += 8 * dt
        if self.wobbling:
            self.amplitude += self.amplitude_accel * dt
            if self.amplitude > self.amp_limit:
                self.amplitude = self.amp_limit
        else:
            self.amplitude -= self.amplitude_deccel * dt
            if self.amplitude < 0: self.amplitude = 0
        
        self.vel[1] = self.amplitude * 1000 * math.sin(self.angle) * dt
        self.vel[0] = self.init_vel + 8 * self.amplitude

        self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt

        if self.x > const.WIDTH:
            self.done = True
            if self.current_path: self.current_path.delete()

        self.path_points.append([self.x, self.y])
        if len(self.path_points) >= 2:
            self.path_batch.add(2, pyglet.gl.GL_LINE_STRIP, None,
                       ('v2f', self.path_points[-2] + self.path_points[-1]),
                       ('c3B', self.ok_color * 2))

        self.predict(self.wobbling)
        self.predict(not self.wobbling, True)

    def predict(self, wobbling, other=False):
        dt = const.INV_FPS
        x = self.x
        y = self.y
        angle = self.angle
        amplitude = self.amplitude
        vel = list(self.vel)
        points = []
        if other and not wobbling: amplitude *= 0.5
        while x < self.x + 300 and x < const.WIDTH:
            if amplitude > const.EPS:
                angle += 8 * dt
            if wobbling:
                amplitude += self.amplitude_accel * dt
                if amplitude > self.amp_limit:
                    amplitude = self.amp_limit
            else:
                amplitude -= self.amplitude_deccel * dt
                if amplitude < 0: amplitude = 0
            
            vel[1] = amplitude * 1000 * math.sin(angle) * dt
            vel[0] = self.init_vel + 8 * amplitude

            x += vel[0] * dt
            y += vel[1] * dt

            points += [x, y]

        if not points: return
        
        if not other:
            batch = self.path_batch
            color = self.path_color
            if self.current_path: self.current_path.delete()
            self.current_path = batch.add(len(points) / 2,
                                          pyglet.gl.GL_POINTS,
                                          None, ('v2f', points),
                                          ('c3B', color * (len(points) / 2)))
        else:
            batch = self.other_path_batch
            color = self.other_color
            if self.other_path: self.other_path.delete()
            self.other_path = batch.add(len(points) / 2,
                                        pyglet.gl.GL_POINTS,
                                        None, ('v2f', points),
                                        ('c3B', color * (len(points) / 2)))
    
    def initial_path(self):
        self.current_path = self.path_batch.add(2, pyglet.gl.GL_POINTS, None,
                                           ('v2f', (self.x, self.y,
                                                    const.WIDTH, self.y)),
                                           ('c3B', self.path_color * 2))

    def charge(self):
        self.charges += 1
        if self.charges == 1: self.image = self.charge1
        elif self.charges == 2: self.image = self.charge2
        else: self.image = self.charge3


class Enemy(pyglet.sprite.Sprite):

    def __init__(self, batch, pos):
        image = pyglet.resource.image('enemy-anim.png')
        img_seq = pyglet.image.ImageGrid(image, 1, 4)
        self.image_normal = img_seq[0]
        self.image_dying = pyglet.image.Animation.from_image_sequence(
            img_seq[1:4], 0.3, False)
        anchor_x = self.image_normal.width / 2
        anchor_y = self.image_normal.height / 2
        self.image_normal.anchor_x = anchor_x
        self.image_normal.anchor_y = anchor_y
        for frame in self.image_dying.frames:
            frame.image.anchor_x = anchor_x
            frame.image.anchor_y = anchor_y
        self.image_dying.anchor_x = anchor_x
        self.image_dying.anchor_y = anchor_y
        super(Enemy, self).__init__(self.image_normal, batch=batch)

        self.x = pos[0]
        self.y = pos[1]

        self.vel = [-90, 0]

        self.done = False
        self.destroyed = False
        self.passed = False

        self.destroyed_timer = 0

    def destroy(self):
        self.destroyed = True
        self.image = self.image_dying
        self.destroyed_timer = self.image_dying.get_duration()
    
    def update(self, dt):
        self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt

        if self.x < -self.width and not self.destroyed:
            self.done = True
            self.passed = True

        if self.destroyed:
            self.destroyed_timer -= dt
            if self.destroyed_timer <= 0:
                self.done = True


class BulletGenerator:

    def __init__(self, batch, life_mgr):
        self.bullets = [None, None, None]
        self.batch = batch
        self.life_mgr = life_mgr

        self.generate(0)
        self.generate(1)
        self.generate(2)

        self.spawn_timer = [0, 0, 0]
        self.spawn_timer_max = 2

    def generate(self, number):
        if self.bullets[number]: self.bullets[number].delete()
        self.bullets[number] = Bullet(self.batch, (0, 120 + number * 120))
    
    def toggle_wobbling(self, number):
        if self.bullets[number]:
            if self.bullets[number].wobbling:
                self.bullets[number].stop_wobbling()
            else:
                self.bullets[number].start_wobbling()

    def draw(self):
        for bullet in self.bullets:
            if bullet:
                bullet.other_path_batch.draw()            
                bullet.path_batch.draw()

    def update(self, dt):
        for i in [0, 1, 2]:
            if self.bullets[i]:
                self.bullets[i].update(dt)
                if self.bullets[i].done:
                    if self.bullets[i].charge > 0:
                        self.life_mgr.enemy_lose(self.bullets[i].charges)
                    self.bullets[i].reset((0, 120 + i * 120))
            else:
                self.spawn_timer[i] += dt
                if self.spawn_timer[i] >= self.spawn_timer_max:
                    self.spawn_timer[i] = 0
                    self.generate(i)


class EnemyGenerator:

    def __init__(self, batch, life_mgr):
        self.enemies = []
        self.batch = batch
        self.life_mgr = life_mgr

        self.enemy_timer_max = 4
        self.enemy_timer = 0

    def generate(self):
        enemy = Enemy(self.batch, (const.WIDTH,
                                   random.randint(80, const.HEIGHT - 80)))
        self.enemies.append(enemy)

    def update(self, dt):
        self.enemy_timer += dt
        if self.enemy_timer >= self.enemy_timer_max:
            self.enemy_timer = 0
            self.generate()
        remove_list = []
        for enemy in self.enemies:
            enemy.update(dt)
            if enemy.done:
                remove_list.append(enemy)
                if enemy.passed: self.life_mgr.player_lose()
        for enemy in remove_list:
            self.enemies.remove(enemy)
            enemy.delete()


class Charge(pyglet.sprite.Sprite):

    def __init__(self, batch, pos):
        image = pyglet.resource.image('charge-anim.png')
        img_seq = pyglet.image.ImageGrid(image, 2, 4)
        self.first_image = pyglet.image.Animation.from_image_sequence(
            img_seq[4:5], 0.1, True)
        self.second_image = pyglet.image.Animation.from_image_sequence(
            img_seq[5:7], 0.1, True)
        self.full_image = pyglet.image.Animation.from_image_sequence(
            img_seq[0:4], 0.1, True)
        for frame in (self.first_image.frames + self.second_image.frames +
                      self.full_image.frames):
            frame.image.anchor_x = frame.image.anchor_y = 8
        self.first_image.anchor_x = self.first_image.anchor_y = 8
        self.second_image.anchor_x = self.second_image.anchor_y = 8
        self.full_image.anchor_x = self.full_image.anchor_y = 8        
        super(Charge, self).__init__(self.first_image, batch=batch)

        self.x = pos[0]
        self.y = pos[1]

        self.timer = 10
        self.done = False

    def update(self, dt):
        self.timer -= dt
        # This is wrong, I know, but it produced a pretty cool effect. :)
        if self.timer >= 9.5 and self.image != self.first_image:
            self.image = self.first_image
        elif self.timer >= 8 and self.image != self.second_image:
            self.image = self.second_image
        elif self.timer >= 2 and self.image != self.full_image:
            self.image = self.full_image
        elif self.timer >= 0.5 and self.image != self.second_image:
            self.image = self.second_image
        elif self.timer >= 0 and self.image != self.first_image:
            self.image = self.first_image
        else:
            self.done = True


class ChargeGenerator:

    def __init__(self, batch):
        self.charges = []
        self.batch = batch

        self.charge_limit = 10
        self.charge_timer_max = 3
        self.charge_timer = self.charge_timer_max
    
    def generate(self):
        pos = (random.randint(90, const.WIDTH - 90),
               random.randint(60, const.HEIGHT - 60))
        charge = Charge(self.batch, pos)
        self.charges.append(charge)

    def update(self, dt):
        self.charge_timer += dt
        if self.charge_timer >= self.charge_timer_max:
            self.charge_timer = 0
            if len(self.charges) < self.charge_limit:
                self.generate()
                self.generate()
        remove_list = []
        for charge in self.charges:
            charge.update(dt)
            if charge.done: remove_list.append(charge)
        for charge in remove_list:
            self.charges.remove(charge)
            charge.delete()


class CollisionManager:

    def __init__(self, bullets, enemies, charges):
        self.bullets = bullets
        self.enemies = enemies
        self.charges = charges

    def update(self, dt):
        for bullet in self.bullets:
            if bullet:
                for charge in self.charges:
                    if (charge and self.hit_sprite(bullet, charge) and
                        bullet.charges < 3):
                        bullet.charge()
                        charge.done = True
                for enemy in self.enemies:
                    if enemy and self.hit_sprite(bullet, enemy):
                        enemy.destroy()

    def hit_sprite(self, sprite1, sprite2):
        s1x = sprite1.x - sprite1.image.anchor_x
        s2x = sprite2.x - sprite2.image.anchor_x
        s1y = sprite1.y - sprite1.image.anchor_y
        s2y = sprite2.y - sprite2.image.anchor_y
        return (not (s1x > s2x + sprite2.width or
                     s1x + sprite1.width < s2x or
                     s1y + sprite1.height < s2y or
                     s1y > s2y + sprite2.height))


class Life(pyglet.sprite.Sprite):

    def __init__(self, batch, pos):
        image = pyglet.resource.image('life.png')
        image.anchor_x = image.width / 2
        image.anchor_y = image.height / 2
        super(Life, self).__init__(image, batch=batch)

        self.x = pos[0]
        self.y = pos[1]
        
        self.done = False


class LifeManager:

    def __init__(self, batch):
        self.batch = batch
        
        self.player_lives = 6
        self.enemy_lives = 16

        self.player_lives_spr = []
        self.enemy_lives_spr = []

        hh = const.HEIGHT / 2
        for i in xrange(self.player_lives):
            offset_x = 10
            if i % 2 == 0:
                pos = (offset_x, hh + (i / 2) * 10)
            else:
                pos = (offset_x, hh - ((i / 2) + 1) * 10)
                
            self.player_lives_spr.append(Life(self.batch, pos))

        for i in xrange(self.enemy_lives):
            offset_x = 10
            if i % 2 == 0:
                pos = (const.WIDTH - offset_x, hh + (i / 2) * 10)
            else:
                pos = (const.WIDTH - offset_x, hh - ((i / 2) + 1) * 10)
                
            self.enemy_lives_spr.append(Life(self.batch, pos))

        self.end_code = 0
        self.message = None
        self.message_timer = 0

    def player_lose(self):
        self.player_lives -= 1
        self.player_lives_spr[-1].delete()
        self.player_lives_spr.pop()
        if self.player_lives <= 0:
            self.end_code = -1
        else:
            if self.message: self.message.delete()
            self.message = pyglet.text.Label(
                'The enemy passed your defenses! You lose a life!',
                batch=self.batch,
                x=const.WIDTH / 2, y=const.HEIGHT-20, font_size=14,
                anchor_x='center', anchor_y='center', color=(200,0,0,255))
            self.message_timer = 4
    
    def enemy_lose(self, number):
        real_number = number
        if number == 3: number = 4 #Bad place to put this, I know. In a hurry.
        self.enemy_lives -= number
        if self.enemy_lives < 0: number += self.enemy_lives
        for i in xrange(number):
            self.enemy_lives_spr[-1].delete()
            self.enemy_lives_spr.pop()
        if self.enemy_lives <= 0:
            self.end_code = 1
        else:
            if self.message: self.message.delete()
            if real_number == 1:
                text = 'Single-charged orb attack! Enemy loses 1 life!'
            elif real_number == 2:
                text = 'Double-charged orb attack!! Enemy loses 2 lives!'
            elif real_number == 3:
                text = 'TRIPLE-charged orb attack!!! Enemy loses 4 lives!'
            else:
                text = ''
            self.message = pyglet.text.Label(
                text, batch=self.batch,
                x=const.WIDTH / 2, y=const.HEIGHT-20, font_size=14,
                anchor_x='center', anchor_y='center', color=(0,200,0,255))
            self.message_timer = 4

    def update(self, dt):
        if self.message:
            self.message_timer -= dt
            if self.message_timer <= 0:
                self.message_timer = 0
                self.message.delete()
                self.message = None
